#define STRICT
#include "../include/os.h"

/* Compile this module only for Windows. */
#ifdef __MSW__
#include <stdio.h>

#include <windows.h>	// Header File For Windows
#include <gl\gl.h>	// Header File For The OpenGL32 Library
#include <gl\glu.h>	// Header File For The GLu32 Library
#include <gl\glaux.h>	// Header File For The Glaux Library

#include "gw.h"
#include "stategl.h"


long GWWGetCurMilliTime(void);
int GWStrCaseCmp(const char *s1, const char *s2);
Boolean GWParseGeometry(
	const char *geometry,
	int *x, int *y, int *width, int *height
);
int GWGetKeyValueFromWParam(gw_display_struct *display, WPARAM wParam);
void GWSetWindowIconFile(
        gw_display_struct *dpy, int ctx_num,
        const char *icon_path, const char *icon_name
);
int GWCreateWindow(
	gw_display_struct *dpy,
	void *parent,
	void **hwnd_rtn,
	void **dc_rtn,
	void **rc_rtn,
	int x, int y,
	int width, int height,
	int depth,		/* In bits. */
	const char *title
);

gw_display_struct *GWInit(int argc, char **argv);
LRESULT CALLBACK WndProc(
        HWND hWnd,      // Handle For This Window 
        UINT uMsg,      // Message For This Window
        WPARAM wParam,  // Additional Message Information
        LPARAM lParam   // Additional Message Information
);
void GWManage(gw_display_struct *display);
void GWShutdown(gw_display_struct *display);

void GWOutputMessage(
	gw_display_struct *display,
	int type,
	const char *subject,
	const char *message,
	const char *help_details
);
int GWConfirmation(
	gw_display_struct *display,
	int type,
	const char *subject,
        const char *message,
        const char *help_details
);
int GWConfirmationSimple(
	gw_display_struct *display,
	const char *message
);

void GWSetDrawCB(
	gw_display_struct *display,
	void (*func)(int, void *),
	void *data
);
void GWSetResizeCB(
	gw_display_struct *display,
        void (*func)(int, void *, int, int, int, int),
        void *data
);
void GWSetKeyboardCB(
	gw_display_struct *display,
	void (*func)(void *, int, Boolean),
	void *data
);
void GWSetPointerCB(
	gw_display_struct *display,
	void (*func)(int, void *, int, int, int),
	void *data
);
void GWSetVisibilityCB(
	gw_display_struct *display,
	void (*func)(int, void *, int),
	void *data
);
void GWSetSaveYourselfCB(
        gw_display_struct *display,
        void (*func)(int, void *),
        void *data
);
void GWSetCloseCB(
	gw_display_struct *display,
	void (*func)(int, void *, void *),
	void *data
);
void GWSetTimeoutCB(
	gw_display_struct *display,
	void (*func)(void *),
        void *data
);

int GWContextNew(
        gw_display_struct *display,
        int x, int y, int width, int height,
        const char *title,
        const char *icon_path, const char *icon_name,
        Bool no_windows
);
void GWContextDelete(
        gw_display_struct *display, int ctx_num
);
int GWContextGet(
        gw_display_struct *display, int ctx_num,
        void **window_id_rtn, void **gl_context_rtn,
        int *x_rtn, int *y_rtn, int *width_rtn, int *height_rtn
);
int GWContextSet(gw_display_struct *display, int ctx_num);
void GWPostRedraw(gw_display_struct *display);
void GWSwapBuffer(gw_display_struct *display);

void GWKeyboardAutoRepeat(gw_display_struct *display, Boolean b);

void GWSetPointerCursor(
        gw_display_struct *display,
        int cursor      /* One of GWPointerCursor*. */
);

void GWSetInputBusy(gw_display_struct *display);
void GWSetInputReady(gw_display_struct *display);

void GWOrtho2D(gw_display_struct *display);



#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))

/*
 * Array of 256 unsigned bytes for recording key states.
 */
#define GW_KEY_STATE_TABLE_MAX_KEYS	256
static unsigned char key_state[256];

/*
 *	Windows key value to GW key value lookup table.
 *
 *	Used in GWGetKeyValueFromWParam(), even number index values
 *	are Windows key values and odd number index values are
 *	GW key values.
 */
#define WINDOWS_TO_GW_KEYTABLE	\
{ \
	VK_MENU, GWKeyAlt, \
	VK_LMENU, GWKeyAlt, \
	VK_RMENU, GWKeyAlt, \
	VK_CONTROL, GWKeyCtrl, \
	VK_LCONTROL, GWKeyCtrl, \
	VK_RCONTROL, GWKeyCtrl, \
	VK_SHIFT, GWKeyShift, \
	VK_LSHIFT, GWKeyShift, \
	VK_RSHIFT, GWKeyShift, \
 \
	VK_NUMPAD0, GWKeyInsert, \
	VK_NUMPAD1, GWKeyEnd, \
	VK_NUMPAD2, GWKeyDown, \
	VK_NUMPAD3, GWKeyPageDown, \
	VK_NUMPAD4, GWKeyLeft, \
	VK_NUMPAD5, GWKeyCenter, \
	VK_NUMPAD6, GWKeyRight, \
	VK_NUMPAD7, GWKeyHome, \
	VK_NUMPAD8, GWKeyUp, \
	VK_NUMPAD9, GWKeyPageUp, \
	VK_MULTIPLY, '*', \
	VK_ADD, '+', \
	VK_SEPARATOR, '=', \
	VK_SUBTRACT, '-', \
	VK_DECIMAL, '.', \
	VK_DIVIDE, '/', \
 \
	VK_F1, GWKeyF1, \
	VK_F2, GWKeyF2, \
	VK_F3, GWKeyF3, \
	VK_F4, GWKeyF4, \
	VK_F5, GWKeyF5, \
	VK_F6, GWKeyF6, \
	VK_F7, GWKeyF7, \
	VK_F8, GWKeyF8, \
	VK_F9, GWKeyF9, \
	VK_F10, GWKeyF10, \
	VK_F11, GWKeyF11, \
	VK_F12, GWKeyF12, \
	VK_F13, GWKeyF13, \
	VK_F14, GWKeyF14, \
	VK_F15, GWKeyF15, \
	VK_F16, GWKeyF16, \
	VK_F17, GWKeyF17, \
	VK_F18, GWKeyF18, \
	VK_F19, GWKeyF19, \
	VK_F20, GWKeyF20, \
 \
	VK_SPACE, ' ', \
	VK_ESCAPE, 0x1b, \
	VK_UP, GWKeyUp, \
	VK_DOWN, GWKeyDown, \
	VK_LEFT, GWKeyLeft, \
	VK_RIGHT, GWKeyRight, \
/*	VK_CENTER, GWKeyCenter, */ \
	VK_HOME, GWKeyHome, \
	VK_END, GWKeyEnd, \
	VK_PRIOR, GWKeyPageUp, \
	VK_NEXT, GWKeyPageDown, \
 \
	VK_TAB, '\t', \
	VK_RETURN, '\n', \
	VK_BACK, GWKeyBackSpace, \
	VK_DELETE, GWKeyDelete, \
	VK_INSERT, GWKeyInsert, \
 \
	VK_PAUSE, GWKeyPause, \
	VK_SCROLL, GWKeyScrollLock, \
	VK_PRINT, GWKeySysReq, \
 \
/* No VK_* definations in Windows for these values, match as ASCII values. */ \
	(unsigned char)'.', '.', \
	(unsigned char)'\'', '\'', \
	(unsigned char)',', ',', \
	(unsigned char)'-', '-', \
	(unsigned char)'_', '_', \
	(unsigned char)'+', '+', \
	(unsigned char)'=', '=', \
	(unsigned char)'/', '/', \
 \
	0, 0	/* Windows key value of 0 marks end of table. */ \
}



/*
 *	Returns the current time in milliseconds, used for time
 *	stamps on event message returns.
 */

long GWWGetCurMilliTime(void)
{
        SYSTEMTIME t;   /* Current time of day structure. */

        GetSystemTime(&t);
        return(
            (time_t)(
                (((((t.wHour * 60.0) + t.wMinute) * 60) + t.wSecond) * 1000) +
                t.wMilliseconds
            )
        );
}

/*
 *	Basically strcasecmp.
 */
int GWStrCaseCmp(const char *s1, const char *s2)
{
	if((s1 == NULL) ||
           (s2 == NULL)
	)
	    return(1);	/* False. */

	while((*s1) && (*s2))
	{
	    if(toupper(*s1) != toupper(*s2))
		return(1);	/* False. */

	    s1++;
	    s2++;
	}
	if((*s1) == (*s2))
	    return(0);	/* True. */
	else
	    return(1);	/* False. */
}

/*
 *	Parses the given geometry string which is of the format
 *	"WxH+X+Y" (variables are uppercase). Alternate geometry
 *	string format "WxH" is also acceptable and the returns for
 *	x and y will be set to 0 and 0 (respectivly) in such case.
 *
 *	Returns True on success and False on error.
 */
Boolean GWParseGeometry(
	const char *geometry,
	int *x, int *y, int *width, int *height
)
{
	const char *s = geometry;

	if(x != NULL)
	    (*x) = 0;
	if(y != NULL)
	    (*y) = 0;
	if(width != NULL)
	    (*width) = 0;
	if(height != NULL)
	    (*height) = 0;

	if(s == NULL)
	    return(False);

	/* Get width. */
	if(width != NULL)
	    (*width) = atoi(s);

	/* Seek to dimensions deliminator 'x'. */
	s = (const char *)strchr(s, 'x');
	if(s == NULL)
	{
	    /* Must have complete dimensions, so this would be an
	     * error.
	     */
	    return(False);
	}
	else
	{
	    /* Seek past deliminator. */
	    s += 1;
	}

	/* Get height. */
	if(height != NULL)
	    (*height) = atoi(s);

	/* Now check if coordinates are given. */
	while(((*s) != '\0') && ((*s) != '+') && ((*s) != '-'))
	    s++;
	if((*s) == '\0')
	{
	    /* No coordinates in the given geometry string,
	     * return success though.
	     */
	    return(True);
	}
	else
	{
	    /* Seek past deliminator (but not if its '-'). */
	    if((*s) == '+')
		s += 1;
	}

	/* Get x coordinate. */
	if(x != NULL)
	    (*x) = atoi(s);

	/* Seek to next coordinate. */
	s += 1;
	while(((*s) != '\0') && ((*s) != '+') && ((*s) != '-'))
	    s++;
	if((*s) == '\0')
	{
	    /* Second coordinate not given, this is an error. */
	    return(False);
	}
	else
	{
	    /* Seek past deliminator (but not if its '-'). */
	    if((*s) == '+')
		s += 1;
	}

	/* Get y coordinate. */
	if(y != NULL)
	    (*y) = atoi(s);

	return(True);
}

/*
 *	Returns a GW key value from the given wParam.
 */
int GWGetKeyValueFromWParam(gw_display_struct *display, WPARAM wParam)
{
	const unsigned int keytable[] = WINDOWS_TO_GW_KEYTABLE;
	const unsigned int *kt_ptr = &(keytable[0]);


	if(display == NULL)
	    return('\0');

	while((*kt_ptr) != 0)
	{
	    if(*kt_ptr++ == wParam)
		return((int)*kt_ptr);

	    kt_ptr++;
	}

	/* No match, return as character literal (make sure lower case). */
	return(tolower((int)wParam));
}


/*
 *	Loads an icon for the toplevel window (does nothing for now).
 */
void GWSetWindowIconFile(
	gw_display_struct *dpy, int ctx_num,
	const char *icon_path, const char *icon_name
)
{
	return;
}


/*
 *	Creates a new toplevel window and sets it up for OpenGL rendering.
 *
 *	All inputs must be valid unless otherwise indicated.
 *
 *	Returns 0 on success or -1 on error.
 */
int GWCreateWindow(
	gw_display_struct *dpy,
	void *parent,		/* Can be NULL to indicate root. */
	void **hwnd_rtn,
	void **dc_rtn,
	void **rc_rtn,
	int x, int y,
	int width, int height,
	int depth,		/* In bits. */
	const char *title
)
{
	void *w, *dc, *rc;	/* Window, device context, and rendering context. */
	GLuint PixelFormat;	// Holds The Results After Searching For A Match
	WNDCLASS wc;		// Windows Class Structure
	DWORD dwExStyle;	// Window Extended Style
	DWORD dwStyle;	// Window Style
	RECT WindowRect;	// Grabs Rectangle Upper Left / Lower Right Values


	if(hwnd_rtn != NULL)
	    *hwnd_rtn = NULL;
	if(dc_rtn != NULL)
	    *dc_rtn = NULL;
	if(rc_rtn != NULL)
	    *rc_rtn = NULL;

	if((hwnd_rtn == NULL) || (dc_rtn == NULL) || (rc_rtn == NULL))
	    return(-1);

	/* Set window size (left and top are not positions). */
	WindowRect.left = (long)0;
	WindowRect.right = (long)width;
	WindowRect.top = (long)0;
	WindowRect.bottom = (long)height;

	/* Begin setting up window class. */
	/* Redraw on resize and own dc for the window. */
	wc.style = CS_HREDRAW | CS_VREDRAW | CS_OWNDC;

	/* Set WndProc() as callback for message (event) handling. */
	wc.lpfnWndProc = (WNDPROC)WndProc;

	/* No extra window class data. */
	wc.cbClsExtra = 0;

	/* Extra window data for client data. */
	wc.cbWndExtra = 4;

	/* Set the instance. */
	wc.hInstance = (HINSTANCE)dpy->pid;

	/* Load default icon. */
	wc.hIcon = LoadIcon(NULL, IDI_WINLOGO);

	/* Load arrow pointer. */
	wc.hCursor = LoadCursor(NULL, IDC_ARROW);

	/* No background. */
	wc.hbrBackground = NULL;

	/* No menu. */
	wc.lpszMenuName = NULL;

	/* Class name. */
	wc.lpszClassName = "OpenGL";

	/* Try to register the window class. */
	if(!RegisterClass(&wc))
	{
	    MessageBox(
		NULL,
		"Failed To Register The Window Class.",
		"ERROR",
		MB_OK | MB_ICONEXCLAMATION
	    );
	    return(-1);
	}

	/* Use full screen mode? */
	if(dpy->fullscreen)
	{
	    /* Set up device mode (for specifying full screen
	     * parameters).
	     */
	    DEVMODE dmScreenSettings;

	    /* Reset device mode structure values. */
	    memset(
		&dmScreenSettings,
		0x00,
		sizeof(dmScreenSettings)
	    );

	    /* Set values for device mode structure. */
	    dmScreenSettings.dmSize = sizeof(dmScreenSettings);

	    dmScreenSettings.dmPelsWidth = width;
	    dmScreenSettings.dmPelsHeight = height;
	    dmScreenSettings.dmBitsPerPel = depth;

	    dmScreenSettings.dmFields = DM_BITSPERPEL | DM_PELSWIDTH |
	                                DM_PELSHEIGHT;

	    /* Try To Set Selected Mode And Get Results.
	     * NOTE: CDS_FULLSCREEN Gets Rid Of Start Bar.
	     */
	    if(ChangeDisplaySettings(
		&dmScreenSettings,
		CDS_FULLSCREEN
	    ) != DISP_CHANGE_SUCCESSFUL)
	    {
		/* If The Mode Fails, Offer Two Options.  Quit Or Use
		 * Windowed Mode.
		 */
		if(MessageBox(
		    NULL,
"The requested fullscreen mode is not supported by\n\
your video card. Use Windowed Mode instead?",
		    "OpenGL",
		    MB_YESNO | MB_ICONEXCLAMATION
		) == IDYES)
		{
		    /* Windowed Mode selected. */
		    dpy->fullscreen = FALSE;
		}
		else
		{
		    /* User does not want windowed mode. */
		    MessageBox(
			NULL,
			"The application will now be closed.",
			"ERROR",
			MB_OK | MB_ICONSTOP
		    );
		    return(-1);
		}
	    }
	}

	/* Still in fullscreen mode? */
	if(dpy->fullscreen)
	{
	    dwExStyle = WS_EX_APPWINDOW;	// Window Extended Style
	    dwStyle = WS_POPUP;			// Windows Style

	    /* Hide cursor while in full screen mode? */
	    if(dpy->fullscreen_hide_cursor)
		ShowCursor(FALSE);
	}
	else
	{
	    /* Window extended style. */
	    dwExStyle = WS_EX_APPWINDOW | WS_EX_WINDOWEDGE;
	    dwStyle = WS_OVERLAPPEDWINDOW;	// Windows Style
	}

	/* Adjust window to actual requested size. */
	AdjustWindowRectEx(&WindowRect, dwStyle, FALSE, dwExStyle);


	/* Create the window. */
	w = CreateWindowEx(
	    dwExStyle,		// Extended Style For The Window
	    "OpenGL",		// Class Name
	    title,		// Window Title
	    dwStyle |		// Defined Window Style
	    WS_CLIPSIBLINGS |	// Required Window Style
	    WS_CLIPCHILDREN,	// Required Window Style
	    x, y,		// Window Position
	    WindowRect.right-WindowRect.left,	// Calculate Window Width
	    WindowRect.bottom-WindowRect.top,	// Calculate Window Height
	    NULL,		// No Parent Window
	    NULL,		// No Menu
	    (HINSTANCE)dpy->pid,	// Instance
	    dpy			/* Client data. */
	);
	if(!w)
	{
	    MessageBox(
		NULL,
		"Window Creation Error.",
		"ERROR",
		MB_OK | MB_ICONEXCLAMATION
	    );
	    return(-1);
	}
	if(hwnd_rtn != NULL)
	    *hwnd_rtn = w;


	/* Set up pixel format descriptor, this defines the per
	 * pixel specifications.
	 */
	static PIXELFORMATDESCRIPTOR pfd =
	{
	    sizeof(PIXELFORMATDESCRIPTOR),	/* Size of this structure. */
	    1,				/* Version number. */
	    PFD_DRAW_TO_WINDOW | 	/* Format must support window. */
	    PFD_SUPPORT_OPENGL |	/* Format must support OpenGL. */
	    PFD_DOUBLEBUFFER,		/* Must support double buffer. */
	    PFD_TYPE_RGBA,		/* Request RGBA format. */
	    ((depth > 16) ? 24 : depth),	/* Color depth. */
	    0, 0, 0, 0, 0, 0,		/* Ignore colors bits. */
	    8,				/* Alpha bits. */
	    0,				/* Alpha shift. */
	    0,				/* Accumulation buffer bits. */
	    0, 0, 0, 0,			/* Accumulation bits (RGBA). */
	    16,				/* Z depth buffer bits. */
	    8,				/* Stencil buffer bits. */
	    0,				/* Auxiliary buffer bits. */
	    PFD_MAIN_PLANE,		/* Main drawing layer. */
	    0,				/* Reserved. */
	    0, 0, 0			/* Layer masks ignored. */
	};

	/* Create the device context. */
	dc = GetDC((HWND)w);
	if(dc == NULL)
	{
	    MessageBox(
		NULL,
		"Can't Create A GL Device Context.",
		"ERROR",
		MB_OK | MB_ICONEXCLAMATION
	    );
	    return(-1);
	}
	if(dc_rtn != NULL)
	    *dc_rtn = dc;


	/* Select pixel format descriptor. */
	PixelFormat = ChoosePixelFormat((HDC)dc, &pfd);
	if(PixelFormat)
	{
	    dpy->has_double_buffer = True;
	    dpy->alpha_channel_bits = 8;
	}
	else
	{
	    dpy->has_double_buffer = False;
	    dpy->alpha_channel_bits = 0;
	    MessageBox(
		NULL,
		"Cannot Find A Suitable PixelFormat.",
		"ERROR",
		MB_OK | MB_ICONEXCLAMATION
	    );
	    return(-1);
	}
/* Need to try other pixel formats? */

	/* Able to set pixel format? */
	if(!SetPixelFormat((HDC)dc, PixelFormat, &pfd))
	{
	    MessageBox(
		NULL,
		"Can't Set The PixelFormat.",
		"ERROR",
		MB_OK | MB_ICONEXCLAMATION
	    );
	    return(-1);
	}

	/* Create rendering context from the hardware device context. */
	rc = wglCreateContext((HDC)dc);
	if(rc == NULL)
	{
	    MessageBox(
		NULL,
		"Can't Create A GL Rendering Context.",
		"ERROR",
		MB_OK | MB_ICONEXCLAMATION
	    );
	    return(-1);
	}
	if(rc_rtn != NULL)
	    *rc_rtn = rc;

	ShowWindow((HWND)w, SW_SHOW);	/* Show the window. */
	SetForegroundWindow((HWND)w);	// Slightly Higher Priority
	SetFocus((HWND)w);			// Sets Keyboard Focus To The Window

	return(0);
}

/*
 *	Initializes the graphics wrapper.
 */
gw_display_struct *GWInit(int argc, char **argv)
{
	int i, ctx_num, no_windows = 0;
	const char *arg_ptr, *cstrptr;
	int x = 0, y = 0;
	int width = 320, height = 240;
	int depth = 16;
	const char *title = "Untitled";
	gw_display_struct *display;
	double aspect_offset = 0.0;
	Boolean fullscreen = False;


	/* Parse arguments. */
	for(i = 0; i < argc; i++)
	{
	    arg_ptr = (const char *)argv[i];
	    if(arg_ptr == NULL)
		continue;

	    /* Window title. */
	    if(!GWStrCaseCmp(arg_ptr, "--title") ||
               !GWStrCaseCmp(arg_ptr, "-title")
	    )
	    {
		i++;
		if(i < argc)
		{
		    arg_ptr = (const char *)argv[i];
		    title = arg_ptr;
		}
		else
		{
		    fprintf(
			stderr,
			"%s: Requires argument.\n",
			argv[i - 1]
		    );
		}
	    }
	    /* Use full screen mode? */
	    else if(!GWStrCaseCmp(arg_ptr, "--fullscreen") ||
		    !GWStrCaseCmp(arg_ptr, "-fullscreen") ||
                    !GWStrCaseCmp(arg_ptr, "--full_screen") ||
                    !GWStrCaseCmp(arg_ptr, "-full_screen")
	    )
	    {
		fullscreen = True;
	    }
	    /* Position and size of toplevel window. */
	    else if(!GWStrCaseCmp(arg_ptr, "--geometry") ||
		    !GWStrCaseCmp(arg_ptr, "-geometry")
	    )
	    {
		i++;
		if(i < argc)
		{
		    arg_ptr = (const char *)argv[i];
		    GWParseGeometry(arg_ptr, &x, &y, &width, &height);
		}
		else
		{
		    fprintf(
			stderr,
			"%s: Requires argument.\n",
			argv[i - 1]
		    );
		}
	    }
	    /* Aspect offset. */
            else if(!GWStrCaseCmp(arg_ptr, "--aspect_offset") ||
                    !GWStrCaseCmp(arg_ptr, "-aspect_offset") ||
                    !GWStrCaseCmp(arg_ptr, "--aspectoffset") ||
                    !GWStrCaseCmp(arg_ptr, "-aspectoffset")
            )
            {    
                i++;
                if(i < argc)
                {
		    aspect_offset = atof(argv[i]);
                }
                else
                {
                    fprintf(
                        stderr,
                        "%s: Requires argument.\n",
                        argv[i - 1]
                    );
                }
            }
            /* No window, this is a special argument to specify that no
             * windows (or dialogs) be created. 
             */
            else if(!GWStrCaseCmp(arg_ptr, "--no_windows") ||
                    !GWStrCaseCmp(arg_ptr, "-no_windows") ||
                    !GWStrCaseCmp(arg_ptr, "--nowindows") ||
                    !GWStrCaseCmp(arg_ptr, "-nowindows")
            )
            {
                no_windows = 1;
            }
	}

	/* Going to full screen? */
	if(fullscreen)
	{
	    /* Need to adjust width to a suitable value. */
	    if(width <= 320)
		width = 320;
	    else if(width <= 640)
		width = 640;
	    else if(width <= 800)
		width = 800;
	    else if(width <= 1024)
		width = 1024;
	    else if(width <= 1152)
		width = 1152;
	    else
		width = 320;

	    /* Set appropriate height. */
	    height = (int)(width * 0.75);
	}



	/* Allocate a new graphics wrapper display structure. */
	display = (gw_display_struct *)calloc(
	    1,
	    sizeof(gw_display_struct)
	);
	if(display == NULL)
            return(display);

	display->toplevel = NULL;
	display->dc = NULL;
	display->rc = NULL;
	display->toplevel_geometry = NULL;
	display->total_gl_contexts = 0;
	display->gl_context_num = -1;

	/* Get HINSTANCE. */
	display->pid = GetModuleHandle(NULL);

	/* Keyboard auto repeat. */
	display->keyboard_autorepeat = True;

	/* Depth (in bits). */
	display->depth = depth;

	/* Set aspect offset. */
	display->aspect_offset = aspect_offset;

	/* Full screen? Note that the call to GWCreateWindow() will
	 * do the actual switching to full screen.
	 */
	display->fullscreen = fullscreen;

	display->fullscreen_hide_cursor = False;

	/* Reset desktop window values. */
	display->root = NULL;
	display->root_width = 0;
	display->root_height = 0;

	/* Clear key state table. */
	memset(
	    &key_state,
	    0x00,
	    GW_KEY_STATE_TABLE_MAX_KEYS * sizeof(unsigned char)
	);


	/* Create the first toplevel window and gl context and set it
	 * into context.
	 */
	ctx_num = GWContextNew(
	    display,
	    0, 0, width, height,
	    title,
	    NULL,	/* Icon path. */
	    NULL,	/* Icon name. */
	    no_windows
	);
	if(ctx_num > -1)
	{
	    GWContextSet(display, ctx_num);
	}
	else
	{
	    /* Failed to create the first context. */
	}


	/* Get GL version. */
	cstrptr = (const char *)glGetString(GL_VERSION);
	if(cstrptr != NULL)
	{
	    char *gl_vs_ptr;
	    char *gl_vs = strdup(cstrptr);
	    if(gl_vs != NULL)
	    {
		/* Deliminate at space which separates version and vendor
		 * name.
		 */
		gl_vs_ptr = strchr(gl_vs, ' ');
		if(gl_vs_ptr != NULL)
		    (*gl_vs_ptr) = '\0';

		gl_vs_ptr = gl_vs;
		if(gl_vs_ptr != NULL)
		{
		    display->gl_version_major = atoi(gl_vs_ptr);
		    gl_vs_ptr = strchr(gl_vs_ptr, '.');
		    if(gl_vs_ptr != NULL)
			gl_vs_ptr++;
		}
                if(gl_vs_ptr != NULL)
                {
                    display->gl_version_minor = atoi(gl_vs_ptr);
                    gl_vs_ptr = strchr(gl_vs_ptr, '.');
                    if(gl_vs_ptr != NULL)
                        gl_vs_ptr++;
                }

		free(gl_vs);
	    }
	}
    
        /* Get maximum texture sizes. */
        if(1)
        {
            GLint v[1];
            glGetIntegerv(GL_MAX_TEXTURE_SIZE, v);
            display->texture_1d_max = display->texture_2d_max =
                display->texture_3d_max = v[0];
        }

        /* Reset all other OpenGL states on record. */
        StateGLResetAll(&display->state_gl);

	/* Reset other OpenGL states. */
/*	glShadeModel(GL_SMOOTH);
	glClearColor(0.0f, 0.0f, 0.0f, 0.5f);
	glClearDepth(1.0f);
	StateGLEnable(&display->state_gl, GL_DEPTH_TEST);
	StateGLDepthFunc(&display->state_gl, GL_LEQUAL);
	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
 */

	/* Set pixel storage format (for rastering font bitmaps). */
        glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
        glPixelStorei(GL_PACK_ALIGNMENT, 1);

	/* Set other GL defaults that we want. Some of these might default
	 * to on but we want them off to improve performance.
	 */
	StateGLDisable(&display->state_gl, GL_DITHER);

	return(display);
}

/*
 *	Windows event management (`processing').
 */
LRESULT CALLBACK WndProc(
	HWND hWnd,	/* Window ID. */
	UINT uMsg,	/* Message for the window. */
	WPARAM wParam,	/* Message values. */
	LPARAM lParam
)
{
	int ctx_num;
	gw_display_struct *dpy = NULL;

#define DO_MATCH_CTXNUM_WITH_HWND	\
{ \
 /* Match context number with the given window id. */ \
 for(ctx_num = 0; ctx_num < dpy->total_gl_contexts; ctx_num++) \
 { \
  if(dpy->toplevel[ctx_num] == (void *)hWnd) \
   break; \
 } \
 if(ctx_num >= dpy->total_gl_contexts) \
  ctx_num = -1; \
}

	/* Handle by message type. */
	switch(uMsg)
	{
	  /* Create message. */
	  case WM_CREATE:
	    /* Get pointer to display structure. */
	    dpy = (gw_display_struct *)(
		((CREATESTRUCT *)lParam)->lpCreateParams
	    );
	    SetWindowLong(hWnd, 0, (LONG)dpy);
/*	    return(FALSE); */
	    break;

	  /* Activate message. */
	  case WM_ACTIVATE:
	    dpy = (gw_display_struct *)GetWindowLong(hWnd, 0);
	    DO_MATCH_CTXNUM_WITH_HWND
	    if((dpy != NULL) && (ctx_num > -1))
	    {
		if(dpy->func_visibility != NULL)
		{
	            /* HIWORD(wParam) if true means program is minimized. */
		    if(HIWORD(wParam))
			dpy->func_visibility(
			    ctx_num,
			    dpy->func_visibility_data,
			    GWVisibilityFullyObscured
			);
		    else
			dpy->func_visibility(
			    ctx_num,
                            dpy->func_visibility_data,
                            GWVisibilityUnobscured
                        );
		}
		return(FALSE);
	    }
	    break;

	  /* Alt-tab. */
	  case WM_KILLFOCUS:
            dpy = (gw_display_struct *)GetWindowLong(hWnd, 0);
	    DO_MATCH_CTXNUM_WITH_HWND
	    if((dpy != NULL) && (ctx_num > -1))
	    {
		/* Is our toplevel window going out of focus and
		 * are we in full screen mode?
		 */
		if(dpy->fullscreen)
                {
		    /* Set toplevel window back to GUI mode. */
                    ChangeDisplaySettings(NULL, 0);
                    ShowCursor(TRUE);

		    dpy->fullscreen = False;
		    return(FALSE);
		}
	    }
	    break;

	  /* Redraw. */
	  case WM_PAINT:
	    dpy = (gw_display_struct *)GetWindowLong(hWnd, 0);
	    DO_MATCH_CTXNUM_WITH_HWND
	    if((dpy != NULL) && (ctx_num > -1))
	    {
		if(dpy->func_draw != NULL)
		    dpy->func_draw(ctx_num, dpy->func_draw_data);
		/* Continue to default handler, do not return(). */
	    }
	    break;

	  /* System command. */
	  case WM_SYSCOMMAND:
	    /* Handle by system call type. */
            dpy = (gw_display_struct *)GetWindowLong(hWnd, 0);
	    switch(wParam)
	    {
	      /* Screen saver wanting to start. */
	      case SC_SCREENSAVE:
		return(FALSE);
	        break;

	      /* Monitor entering power save. */
	      case SC_MONITORPOWER:
		return(FALSE);
		break;
	    }
	    break;

	  /* Close window. */
	  case WM_CLOSE:
            dpy = (gw_display_struct *)GetWindowLong(hWnd, 0);
	    DO_MATCH_CTXNUM_WITH_HWND
	    if((dpy != NULL) && (ctx_num > -1))
	    {
		PostQuitMessage(0);	/* Send WM_QUIT message. */
		return(FALSE);
	    }
	    break;

	  /* Key press. */
	  case WM_KEYDOWN:
	    /* wParam is a 0 to 255 value indicating the key value.
	     */
            dpy = (gw_display_struct *)GetWindowLong(hWnd, 0);
	    DO_MATCH_CTXNUM_WITH_HWND
	    if((dpy != NULL) && (ctx_num > -1))
	    {
		int i, k;
		Boolean actually_changed = False;

		i = (int)wParam;
		if(i >= GW_KEY_STATE_TABLE_MAX_KEYS)
		    i = GW_KEY_STATE_TABLE_MAX_KEYS - 1;
		if(i < 0)
		    i = 0;

		/* Check if keyboard auto repeat is off. */
		if(dpy->keyboard_autorepeat)
		{
		    if(!(key_state[i]))
			actually_changed = True;
		}
		else
		{
		    if(key_state[i])
			return(FALSE);
		    else
			actually_changed = True;
		}
		key_state[i] = 0x01;

		/* Get GW key value from wParam. */
		k = GWGetKeyValueFromWParam(dpy, wParam);

		/* Skip if key is a modifier and state did not actually change. */
		if(!actually_changed)
		{
		    if((k == GWKeyAlt) ||
                       (k == GWKeyCtrl) ||
		       (k == GWKeyShift)
		    )
			return(FALSE);
		}

	        if(dpy->func_keyboard != NULL)
		{
                    long t = (long)GWWGetCurMilliTime();

                    dpy->func_keyboard(
                        dpy->func_keyboard_data,
                        k,
                        True,	/* Key press. */
			t
                    );
		}
		return(FALSE);
	    }
	    break;

	  /* Key release. */
	  case WM_KEYUP:
	    /* wParam is a 0 to 255 value indicating the key value.
	     */
            dpy = (gw_display_struct *)GetWindowLong(hWnd, 0);
	    DO_MATCH_CTXNUM_WITH_HWND
	    if((dpy != NULL) && (ctx_num > -1))
	    {
		int i, k;
		Boolean actually_changed = False;

		i = (int)wParam;
		if(i >= GW_KEY_STATE_TABLE_MAX_KEYS)
		    i = GW_KEY_STATE_TABLE_MAX_KEYS - 1;
		if(i < 0)
		    i = 0;

		/* Check if keyboard auto repeat is off. */
		if(dpy->keyboard_autorepeat)
		{
		    if(key_state[i])
			actually_changed = True;
		}
		else
		{
		    if(key_state[i])
			actually_changed = True;
		    else
			return(FALSE);
		}
		key_state[i] = 0x00;

		/* Get GW key value from wParam. */
		k = GWGetKeyValueFromWParam(dpy, wParam);

		/* Skip if key is a modifier and state did not actually change. */
		if(!actually_changed)
		{
		    if((k == GWKeyAlt) ||
                       (k == GWKeyCtrl) ||
		       (k == GWKeyShift)
		    )
			return(FALSE);
		}

		if(dpy->func_keyboard != NULL)
		{
                    long t = (long)GWWGetCurMilliTime();
 
                    dpy->func_keyboard(
                        dpy->func_keyboard_data,
			k,
			False, t
		    );
		}
		return(FALSE);
	    }
	    break;

	  /* Move. */
	  case WM_MOVE:
	    /* Ignore this. */
	    break;

	  /* Resize. */
	  case WM_SIZE:
	    /* LoWord = Width, HiWord = Height. */
            dpy = (gw_display_struct *)GetWindowLong(hWnd, 0);
	    DO_MATCH_CTXNUM_WITH_HWND
	    if((dpy != NULL) && (ctx_num > -1))
	    {
		WRectangle *rect = &dpy->toplevel_geometry[ctx_num];

		/* Set quick access geometry values only if this is the
		 * first context.
		 */
		if(ctx_num == 0)
		{
		    dpy->width = LOWORD(lParam);
		    dpy->height = HIWORD(lParam);
		}
		rect->width = LOWORD(lParam);
		rect->height = HIWORD(lParam);

		/* Call resize callback. */
		if(dpy->func_resize != NULL)
		    dpy->func_resize(
			ctx_num, dpy->func_resize_data,
			0, 0,
			rect->width, rect->height
		    );

		/* Call redraw callback as well. */
		if(dpy->func_draw != NULL)
		    dpy->func_draw(ctx_num, dpy->func_draw_data);

		return(FALSE);
	    }
	    break;

	  /* Pointer button press. */
	  case WM_LBUTTONDOWN:
            dpy = (gw_display_struct *)GetWindowLong(hWnd, 0);
	    DO_MATCH_CTXNUM_WITH_HWND
	    if((dpy != NULL) && (ctx_num > -1))
	    {
		if(dpy->func_pointer != NULL)
		{
		    int x = (int)((unsigned int)lParam & 0x0000ffff);
		    int y = (int)(((unsigned int)lParam & 0xffff0000) >> 16);
		    long t = (long)GWWGetCurMilliTime();

		    dpy->func_pointer(
			ctx_num, dpy->func_pointer_data,
			x, y, GWEventTypeButtonPress, 1, t
		    );
		}
		return(FALSE);
	    }
	    break;

          case WM_RBUTTONDOWN:
            dpy = (gw_display_struct *)GetWindowLong(hWnd, 0);
	    DO_MATCH_CTXNUM_WITH_HWND
	    if((dpy != NULL) && (ctx_num > -1))
            {
		if(dpy->func_pointer != NULL)
		{
		    int x = (int)((unsigned int)lParam & 0x0000ffff);
		    int y = (int)(((unsigned int)lParam & 0xffff0000) >> 16);
		    long t = (long)GWWGetCurMilliTime();

		    dpy->func_pointer(
			ctx_num, dpy->func_pointer_data,
			x, y, GWEventTypeButtonPress, 3, t
		    );
		}
		return(FALSE);
	    }
	    break;

	  /* Pointer button release. */
	  case WM_LBUTTONUP:
            dpy = (gw_display_struct *)GetWindowLong(hWnd, 0);
	    DO_MATCH_CTXNUM_WITH_HWND
	    if((dpy != NULL) && (ctx_num > -1))
	    {
		if(dpy->func_pointer != NULL)
		{
		    int x = (int)((unsigned int)lParam & 0x0000ffff);
		    int y = (int)(((unsigned int)lParam & 0xffff0000) >> 16);
		    long t = (long)GWWGetCurMilliTime();

		    dpy->func_pointer(
			ctx_num, dpy->func_pointer_data,
			x, y, GWEventTypeButtonRelease, 1, t
		    );
		}
		return(FALSE);
	    }
	    break;

          case WM_RBUTTONUP:
            dpy = (gw_display_struct *)GetWindowLong(hWnd, 0);
	    DO_MATCH_CTXNUM_WITH_HWND
	    if((dpy != NULL) && (ctx_num > -1))
	    {
		if(dpy->func_pointer != NULL)
		{
		    int x = (int)((unsigned int)lParam & 0x0000ffff);
		    int y = (int)(((unsigned int)lParam & 0xffff0000) >> 16);
		    long t = (long)GWWGetCurMilliTime();

		    dpy->func_pointer(
			ctx_num, dpy->func_pointer_data,
			x, y, GWEventTypeButtonRelease, 3, t
		    );
		}
		return(FALSE);
            }
            break;

	  /* Pointer motion. */
	  case WM_MOUSEMOVE:
            dpy = (gw_display_struct *)GetWindowLong(hWnd, 0);
	    DO_MATCH_CTXNUM_WITH_HWND
	    if((dpy != NULL) && (ctx_num > -1))
	    {
		if(dpy->func_pointer != NULL)
		{
		    int x = (int)((unsigned int)lParam & 0x0000ffff);
		    int y = (int)(((unsigned int)lParam & 0xffff0000) >> 16);
		    long t = (long)GWWGetCurMilliTime();

		    dpy->func_pointer(
			ctx_num, dpy->func_pointer_data,
			x, y, GWEventTypePointerMotion, 0, t
		    );
		}
		return(FALSE);
	    }
	    break;
	}

#undef DO_MATCH_CTXNUM_WITH_HWND

	/* Pass All Unhandled Messages To DefWindowProc */
	return(DefWindowProc(hWnd, uMsg, wParam, lParam));
}

/*
 *	Graphics wrapper management.
 */
void GWManage(gw_display_struct *display)
{
	int ctx_num;
	MSG msg;


	if(display == NULL)
	    return;

#define DO_MATCH_CTXNUM_WITH_HWND	\
{ \
 /* Match context number with the given window id. */ \
 for(ctx_num = 0; ctx_num < display->total_gl_contexts; ctx_num++) \
 { \
  if(display->toplevel[ctx_num] == (void *)hWnd) \
   break; \
 } \
 if(ctx_num >= display->total_gl_contexts) \
  ctx_num = -1; \
}

	/* Update modifier key states. */
	display->alt_key_state = ((GetAsyncKeyState(VK_MENU)) ? True : False);
	display->ctrl_key_state = ((GetAsyncKeyState(VK_CONTROL)) ? True : False);
	display->shift_key_state = ((GetAsyncKeyState(VK_SHIFT)) ? True : False);

	/* Manage event messages. */
	if(PeekMessage(&msg, NULL, 0, 0, PM_REMOVE))
	{
	    HWND hWnd = msg.hwnd;

	    switch(msg.message)
	    {
	      case WM_QUIT:
		if((void *)hWnd == NULL)
		{
		    /* Call close callback for all contexts. */
		    if(display->func_close != NULL)
		    {
			for(ctx_num = 0; ctx_num < display->total_gl_contexts; ctx_num++)
			    display->func_close(
				ctx_num,
				display->func_close_data,
				NULL
			    );
		    }
		}
		else
		{
		    DO_MATCH_CTXNUM_WITH_HWND
		    if((display->func_close != NULL) && (ctx_num > -1))
		    {
			display->func_close(
			    ctx_num,
			    display->func_close_data,
			    (void *)hWnd
			);
		    }
		}
		break;
				      
	      default:
		TranslateMessage(&msg);	// Translate message
		DispatchMessage(&msg);	// Dispatch message
		break;
	    }
	}
	else
	{
	    /* Did not get any new messages, call client timeout function. */
	    if(display->func_timeout != NULL)
		display->func_timeout(
		    display->func_timeout_data
		);
	}

#undef DO_MATCH_CTXNUM_WITH_HWND
}

/*
 *	Graphics wrapper shutdown.
 */
void GWShutdown(gw_display_struct *display)
{
	int i;


        if(display == NULL)
            return;

	/* Are we in full screen mode? */
	if(display->fullscreen)
	{
	    /* Switch back to windowed mode. */
	    ChangeDisplaySettings(NULL, 0);
	    ShowCursor(TRUE);
	}

	if(1)
	{
	    /* Destroy all gl contexts and toplevel Windows, current
	     * gl context will be disabled first as needed.
	     */
	    for(i = display->total_gl_contexts - 1; i >= 0; i--)
		GWContextDelete(display, i);

	    /* Deallocate pointer arrays for all gl contexts and
	     * toplevel Windows.
	     */
	    free(display->toplevel);
	    display->toplevel = NULL;
	    free(display->dc);
	    display->dc = NULL;
	    free(display->rc);
	    display->rc = NULL;
	    free(display->toplevel_geometry);
            display->toplevel_geometry = NULL;
	    display->total_gl_contexts = 0;
	    display->gl_context_num = -1;
	}


	/* Are we able to unregister class? */
	if(!UnregisterClass("OpenGL", (HINSTANCE)display->pid))
	{
	    MessageBox(
		NULL,
		"Could Not Unregister Class.",
		"SHUTDOWN ERROR",
		MB_OK | MB_ICONINFORMATION
	    );
	    display->pid = NULL;
	}

	/* Deallocate display structure. */
	free(display);
}

/*
 *	Output message in dialog form.
 */
void GWOutputMessage(
        gw_display_struct *display,
        int type,       /* One of GWOutputMessageType*. */
        const char *subject,
        const char *message,
        const char *help_details
)
{
	UINT uType = MB_OK | MB_ICONINFORMATION;


	if(display == NULL)
	    return;

	switch(type)
	{
	  case GWOutputMessageTypeGeneral:
	    uType = MB_OK | MB_ICONINFORMATION;
	    break;

	  case GWOutputMessageTypeWarning:
	    uType = MB_OK | MB_ICONWARNING;
	    break;

	  case GWOutputMessageTypeError:
	    uType = MB_OK | MB_ICONERROR;
	    break;

	  case GWOutputMessageTypeQuestion:
            uType = MB_OK | MB_ICONQUESTION;
            break;
	}

	MessageBox(
	    NULL,	/* Window handle. */
	    message,	/* Text. */
	    subject,	/* Caption. */
	    uType
	);
}

/*
 *	Map confirmation dialog and block client execution.
 */
int GWConfirmation(
        gw_display_struct *display,
        int type,       /* One of GWOutputMessageType*. */
        const char *subject,   
        const char *message,
        const char *help_details
)
{
/* Need to work on this. */
	return(GWConfirmationNotAvailable);
}

/*
 *	Simplified version of GWConfirmation().
 */
int GWConfirmationSimple(
        gw_display_struct *display,
        const char *message
)
{
	return(GWConfirmation(
	    display,
	    GWOutputMessageTypeQuestion,
	    "Confirmation",
	    message,
	    NULL
	));
}

/* Set draw callback function, needs to be set before the resize function! */
void GWSetDrawCB(
        gw_display_struct *display,
        void (*func)(int, void *),
	void *data
)
{
	if(display != NULL)
	{
	    display->func_draw = func;
	    display->func_draw_data = data;
	}
}


/* Set resize callback function.

   Reminder, when resize callback is set, a call must be made to it
   immediatly because under Windows, the callback needs
   to be set after the Windows is created and Windows may not send
   a resize request later.

   So this function needs to be set after the draw function.
 */
void GWSetResizeCB(
        gw_display_struct *display, 
        void (*func)(int, void *, int, int, int, int),
        void *data
)
{
	if(display != NULL)
	{
	    display->func_resize = func;
	    display->func_resize_data = data;

	    /* Now call the resize function. */
	    if(display->func_resize != NULL)
		display->func_resize(
		    display->gl_context_num,
		    display->func_resize_data,
		    0, 0,
		    display->width, display->height
		);
	}
}

/* Set keyboard callback function. */
void GWSetKeyboardCB(
        gw_display_struct *display,
        void (*func)(void *, int, Boolean, long),
        void *data
)
{
	if(display != NULL)
	{
	    display->func_keyboard = func;
	    display->func_keyboard_data = data;
	}
}

/* Set pointer callback function. */
void GWSetPointerCB(
	gw_display_struct *display,
        void (*func)(int, void *, int, int, int, int, long),
        void *data
)
{
	if(display != NULL)
	{
	    display->func_pointer = func;
	    display->func_pointer_data = data;
	}
}

/* Set visibility callback function. */
void GWSetVisibilityCB(
        gw_display_struct *display,
        void (*func)(int, void *, int),
        void *data
)
{
	if(display != NULL)
	{
	    display->func_visibility = func;
	    display->func_visibility_data = data;

	    /* Need to send visibility notify upon setting of function,
	     * assume visibility unobscured.
	     */
	    if(display->func_visibility != NULL)
	    {
		display->func_visibility(
		    display->gl_context_num,
		    display->func_visibility_data,
		    GWVisibilityUnobscured
		);
	    }
	}
}

/* Set save yourself callback function. */
void GWSetSaveYourselfCB(
        gw_display_struct *display,
        void (*func)(int, void *),
        void *data
)
{
        if(display != NULL)
	{
            display->func_save_yourself = func;
            display->func_save_yourself_data = data;
	}
}

/* Set close callback function. */
void GWSetCloseCB(
        gw_display_struct *display,
        void (*func)(int, void *, void *),
        void *data
)
{
	if(display != NULL)
	{
	    display->func_close = func;
	    display->func_close_data = data;
	}
}

/* Set timeout callback function. */
void GWSetTimeoutCB(
        gw_display_struct *display,
        void (*func)(void *),
        void *data
)
{
	if(display != NULL)
	{
	    display->func_timeout = func;
	    display->func_timeout_data = data;
	}
}


/*
 *	Creates a new gl context and window, returning the context
 *	number or -1 on error.
 */
int GWContextNew(
        gw_display_struct *display,
        int x, int y, int width, int height,
        const char *title,
        const char *icon_path, const char *icon_name,
        Bool no_windows
)
{
	int ctx_num;
	void *w, *dc, *rc;
	WRectangle *rect;


	if(display == NULL)
	    return(-1);


	/* Search for an available gl context index number. */
	for(ctx_num = 0; ctx_num < display->total_gl_contexts; ctx_num++)
	{
	    w = display->toplevel[ctx_num];
	    dc = display->dc[ctx_num];
	    rc = display->rc[ctx_num];

	    /* This gl context index available? */
	    if((w == NULL) && (dc == NULL) && (rc == NULL))
		break;
	}
	/* No available gl contexts? */
	if(ctx_num >= display->total_gl_contexts)
	{
	    /* Allocate a new gl context. */

	    ctx_num = display->total_gl_contexts;
	    display->total_gl_contexts = ctx_num + 1;

	    display->toplevel = (void **)realloc(
		display->toplevel,
		display->total_gl_contexts * sizeof(void *)
	    );
	    display->dc = (void **)realloc(
		display->dc,
		display->total_gl_contexts * sizeof(void *)
	    );
	    display->rc = (void **)realloc(
		display->rc,
		display->total_gl_contexts * sizeof(void *)
	    );
	    display->toplevel_geometry = (WRectangle *)realloc(
		display->toplevel_geometry,
		display->total_gl_contexts * sizeof(WRectangle)
	    );
	    if((display->toplevel == NULL) || (display->dc == NULL) ||
	       (display->rc == NULL) || (display->toplevel_geometry == NULL)
	    )
	    {
		display->gl_context_num = -1;
		display->total_gl_contexts = 0;
		return(-1);
	    }
	}

	/* At this point we have an available gl context specified by
	 * ctx_num.
	 */


	/* Create first toplevel window and buffer. */
        if(!no_windows)
        {
            /* Set initial geometry of new Window before creation. */
	    rect = &display->toplevel_geometry[ctx_num];

	    /* Set quick access geometry values only if this is the
	     * first context.
	     */
	    if(ctx_num == 0)
	    {
		display->x = x;
		display->y = y;
		display->width = width;
		display->height = height;
	    }

            rect->x = x;
            rect->y = y;
            rect->width = width;
            rect->height = height;

	    /* Create gl rendering context */
	    if(GWCreateWindow(
		display,
		NULL,		/* No parent, since creating a toplevel window. */
		&display->toplevel[ctx_num],
		&display->dc[ctx_num],
		&display->rc[ctx_num],
		x, y, width, height,
		display->depth,	/* Depth in bits. */
		title
	    ))
	    {
		/* Failed to create window. */
		return(-1);
	    }
	}

	return(ctx_num);
}

/*
 *	Destroys the gl context and window given by the context number
 *	ctx_num. If the specified gl context is the current gl context
 *	then the current gl context will be no longer valid.
 */
void GWContextDelete(
        gw_display_struct *display, int ctx_num
)
{
	void *w;
	WRectangle *rect;


	if(display == NULL)
	    return;

	if((ctx_num < 0) || (ctx_num >= display->total_gl_contexts))
	    return;

 	/* Is the given gl context the current gl context? */
	if(ctx_num == display->gl_context_num)
	{
	    display->gl_context_num = -1;

 	    if(!wglMakeCurrent(NULL, NULL))
	    {
		MessageBox(
		    NULL,
		    "Unable To Disable Current GL Context.",
		    "GL Error",
		    MB_OK | MB_ICONERROR
		);
	    }
	}

	/* Release the rendering context. */
	if(display->rc[ctx_num] != NULL)
	{
	    if(!wglDeleteContext((HGLRC)display->rc[ctx_num]))
	    {
		MessageBox(
		    NULL,
		    "Unable To Release Rendering Context.",
		    "GL Error",
		    MB_OK | MB_ICONERROR
		);
	    }
	    display->rc[ctx_num] = NULL;
	}

	/* Release the device context. */
	if((display->toplevel[ctx_num] != NULL) &&
	   (display->dc[ctx_num] != NULL)
	)
	{
	    if(!ReleaseDC(
		(HWND)display->toplevel[ctx_num],
		(HDC)display->dc[ctx_num]
	    ))
	    {
		MessageBox(
		    NULL,
		    "Unable To Release Device Context.",
		    "Windows Error",
		    MB_OK | MB_ICONERROR
		);
	    }
	    display->dc[ctx_num] = NULL;
	}

	/* Destroy the toplevel window. */
	if(display->toplevel[ctx_num] != NULL)
	{
	    if(!DestroyWindow((HWND)display->toplevel[ctx_num]))
	    {
		MessageBox(
		    NULL,
		    "Unable To Destroy Toplevel Window.",
		    "Windows Error",
		    MB_OK | MB_ICONERROR
		);
	    }
	    display->toplevel[ctx_num] = NULL;										// Set hWnd To NULL
	}


	/* Clear the toplevel rectangle. */
	rect = &display->toplevel_geometry[ctx_num];
	memset(rect, 0x00, sizeof(WRectangle));

}


/*
 *	Gets the opaque handle of the Window and the GL rendering
 *	context of the specified GL context.
 *
 *	Returns non-zero on error.
 */
int GWContextGet(
        gw_display_struct *display, int ctx_num,
        void **window_id_rtn, void **gl_context_rtn,
        int *x_rtn, int *y_rtn, int *width_rtn, int *height_rtn
)
{
	void *w;
	void *rc;
	WRectangle *rect;


	if(window_id_rtn != NULL)
	    *window_id_rtn = NULL;
	if(gl_context_rtn != NULL)
	    *gl_context_rtn = NULL;
	if(x_rtn != NULL)
	    *x_rtn = 0;
        if(y_rtn != NULL)
            *y_rtn = 0;
        if(width_rtn != NULL)
            *width_rtn = 0;
        if(height_rtn != NULL)
            *height_rtn = 0;

        if(display == NULL)
            return(-1);

        if((ctx_num < 0) || (ctx_num >= display->total_gl_contexts))
            return(-1);

        w = display->toplevel[ctx_num];
        rc = display->rc[ctx_num];
	rect = &display->toplevel_geometry[ctx_num];

	if((w == NULL) || (rc == NULL))
	    return(-1);


	if(window_id_rtn != NULL)
	    *window_id_rtn = (void *)w;
        if(gl_context_rtn != NULL)
            *gl_context_rtn = (void *)rc;

        if(x_rtn != NULL)
            *x_rtn = (int)rect->x;
        if(y_rtn != NULL)
            *y_rtn = (int)rect->y;
        if(width_rtn != NULL)
            *width_rtn = (int)rect->width;
        if(height_rtn != NULL)
            *height_rtn = (int)rect->height;

	return(0);
}

/*
 *	Puts the GL rendering context specified by ctx_num as the
 *	current rendering GL context.
 *
 *	Returns non-zero on error.
 */
int GWContextSet(gw_display_struct *display, int ctx_num)
{
	void *dc, *rc;


	if(display == NULL)
	    return(-1);

	if((ctx_num < 0) || (ctx_num >= display->total_gl_contexts))
	    return(-1);

        /* Specified gl context already the current gl context? */
        if(ctx_num == display->gl_context_num)
            return(0);


	/* Get device context and rendering context of the specified context. */
	dc = display->dc[ctx_num];
	rc = display->rc[ctx_num];

	if((dc == NULL) || (rc == NULL))
	    return(-1);

	/* Activate the rendering context, thus putting it into GL context. */
	if(!wglMakeCurrent((HDC)dc, (HGLRC)rc))
	{
	    return(-1);
	}
	else
	{
	    display->gl_context_num = ctx_num;
	    return(0);
	}
}

/*
 *	Sends a redraw event (calls redraw callback function immediatly).
 */
void GWPostRedraw(gw_display_struct *display)
{
	int ctx_num;

	if(display == NULL)
	    return;

	ctx_num = display->gl_context_num;
	if((ctx_num < 0) || (ctx_num >= display->total_gl_contexts))
	    return;

	/* Call redraw function. */
	if(display->func_draw != NULL)
	    display->func_draw(
		ctx_num, display->func_draw_data
	    );
}

/*
 *	Swaps buffers.
 */
void GWSwapBuffer(gw_display_struct *display)
{
	int ctx_num;


	if(display == NULL)
	    return;

	ctx_num = display->gl_context_num;
	if((ctx_num < 0) || (ctx_num >= display->total_gl_contexts))
	    return;

	if(display->dc[ctx_num] != NULL)
	    SwapBuffers((HDC)display->dc[ctx_num]);
}

/*
 *	Turns keyboard auto repeat on/off.
 */
void GWKeyboardAutoRepeat(gw_display_struct *display, Boolean b)
{
	if(display != NULL)
	    display->keyboard_autorepeat = b;
}

/*
 *	Sets the pointer cursor to the one specified by cursor.
 */
void GWSetPointerCursor(   
        gw_display_struct *display,
        int cursor      /* One of GWPointerCursor*. */
)
{
        if(display == NULL)
            return;


}

/*
 *	Sets pointer cursor as busy and ignores input.
 */
void GWSetInputBusy(gw_display_struct *display)
{
	if(display == NULL)
	    return;


}

/*
 *      Sets pointer cursor as ready and resumes allowing input.
 */
void GWSetInputReady(gw_display_struct *display)
{
	if(display == NULL)
	    return;


}

/*
 *	Disables depth testing, transforms matrices and model view
 *	for 2d drawing.
 */
void GWOrtho2D(gw_display_struct *display)
{
	int ctx_num;
	WRectangle *rect;


	if(display == NULL)
	    return;

	ctx_num = display->gl_context_num;
	if((ctx_num < 0) || (ctx_num >= display->total_gl_contexts))
	    return;

	rect = &display->toplevel_geometry[ctx_num];

	StateGLDisable(&display->state_gl, GL_DEPTH_TEST);
	StateGLDepthFunc(&display->state_gl, GL_ALWAYS);

        glMatrixMode(GL_PROJECTION);
        glLoadIdentity();
        gluOrtho2D(  
            0, rect->width,
            0, rect->height
        );      
        glMatrixMode(GL_MODELVIEW);   
        glLoadIdentity();
}


#endif	/* __MSW__ */
