#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>

#include "../include/string.h"

#include "sfm.h"
#include "v3dtex.h"

#include "sarreality.h"
#include "obj.h"

#include "sar.h"


void SARObjCreatePremodeledPowerTransmissionTower(
        sar_core_struct *core_ptr, sar_scene_struct *scene,
        int obj_num, sar_object_struct *obj_ptr,
        sar_object_premodeled_struct *obj_premodel_ptr,
        int argc, char **argv
);
void SARObjCreatePremodeledRadioTower(
        sar_core_struct *core_ptr, sar_scene_struct *scene,
        int obj_num, sar_object_struct *obj_ptr,
        sar_object_premodeled_struct *obj_premodel_ptr,
        int argc, char **argv
);
void SARObjCreatePremodeledTower(
        sar_core_struct *core_ptr, sar_scene_struct *scene,
        int obj_num, sar_object_struct *obj_ptr,
        sar_object_premodeled_struct *obj_premodel_ptr,
        int argc, char **argv
);
void SARObjCreatePremodeledControlTower(
        sar_core_struct *core_ptr, sar_scene_struct *scene,
        int obj_num, sar_object_struct *obj_ptr,
        sar_object_premodeled_struct *obj_premodel_ptr,
        int argc, char **argv
);
void SARObjCreatePremodeledBuilding(
        sar_core_struct *core_ptr, sar_scene_struct *scene,
        int obj_num, sar_object_struct *obj_ptr,
        sar_object_premodeled_struct *obj_premodel_ptr,
        int argc, char **argv
);
int SARObjCreatePremodeled(
	sar_core_struct *core_ptr, sar_scene_struct *scene,
	const char *type_name, int argc, char **argv
);


#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))



/*
 *      Creates a power transmission tower.
 *
 *      Inputs assumed valid.
 */
void SARObjCreatePremodeledPowerTransmissionTower(
        sar_core_struct *core_ptr, sar_scene_struct *scene,
        int obj_num, sar_object_struct *obj_ptr,
        sar_object_premodeled_struct *obj_premodel_ptr,
        int argc, char **argv
)
{
	int flashing;
	double z;
        int i, hazard_lights;
	sar_light_struct *light_ptr;


        /* Format:
         *
         * <range> <height> <hazard_lights>
         */

        obj_premodel_ptr->type = SAR_OBJ_PREMODELED_POWER_TRANSMISSION_TOWER;

        /* Range. */
        i = 0;
        obj_ptr->range = ((i < argc) ? atof(argv[i]) : 1000.0);

        /* Height (in feet). */
        i = 1;
        obj_premodel_ptr->height = SFMFeetToMeters(
            (i < argc) ? atof(argv[i]) : 80.0
        );

        /* Number of hazard lights? */
        i = 2;
	hazard_lights = (i < argc) ? atoi(argv[i]) : 0;
	flashing = 1;
	z = obj_premodel_ptr->height;
	for(i = 0; i < hazard_lights; i++)
	{
	    light_ptr = SARObjCreateLight(
		scene,
		&obj_ptr->light, &obj_ptr->total_lights
	    );
	    if(light_ptr != NULL)
	    {
		if(flashing)
		{
		    light_ptr->pos.x = 0.0;
                    light_ptr->pos.y = 0.0;
                    light_ptr->pos.z = z;

                    light_ptr->color.r = 1.0;
                    light_ptr->color.g = 0.0;
                    light_ptr->color.b = 0.0;
                    light_ptr->color.a = 1.0;

                    light_ptr->radius = 2;

		    light_ptr->flags |= SAR_LIGHT_FLAG_ON;
		    light_ptr->flags |= SAR_LIGHT_FLAG_STROBE;

                    light_ptr->int_on = 800;
                    light_ptr->int_off = 1700;
		}
		else
		{
                    light_ptr->pos.x = 0.0;
                    light_ptr->pos.y = 0.0;
                    light_ptr->pos.z = z;

                    light_ptr->color.r = 1.0;
                    light_ptr->color.g = 0.0;
                    light_ptr->color.b = 0.0;
                    light_ptr->color.a = 1.0;

                    light_ptr->radius = 2;

                    light_ptr->flags |= SAR_LIGHT_FLAG_ON;

                    light_ptr->int_on = 0;
                    light_ptr->int_off = 0;
		}
	    }

	    /* Toggle flashing, this is to create alternating steady
	     * state and flashing hazard lights.
	     */
	    flashing = !flashing;

	    /* Decrease height. */
	    z -= (obj_premodel_ptr->height / hazard_lights);
	}

        /* Set cylendrical contact bounds, the radius of the contact
	 * bounds is 0.3 times the height, since the height controls
	 * a uniformed scaling the the unit width is 0.3.
	 */
        SARObjAddContactBoundsCylendrical(
            obj_ptr,
            SAR_CRASH_FLAG_CRASH_CAUSE,
            SAR_CRASH_TYPE_OBSTRUCTION,
	    0.3 * obj_premodel_ptr->height,	/* Radius. */
	    0.0, obj_premodel_ptr->height	/* Height min and max. */
	);
}

/*
 *      Creates a radio tower.
 *
 *      Inputs assumed valid.
 */
void SARObjCreatePremodeledRadioTower(
        sar_core_struct *core_ptr, sar_scene_struct *scene,
        int obj_num, sar_object_struct *obj_ptr,
        sar_object_premodeled_struct *obj_premodel_ptr,
        int argc, char **argv
)
{
        int flashing;
        double z;
        int i, hazard_lights;
        sar_light_struct *light_ptr;


        /* Format:
         *
         * <range> <height> <hazard_lights>
         */

        obj_premodel_ptr->type = SAR_OBJ_PREMODELED_RADIO_TOWER;

        /* Range. */
        i = 0;
        obj_ptr->range = ((i < argc) ? atof(argv[i]) : 1000.0);

        /* Height (in feet). */
        i = 1;
        obj_premodel_ptr->height = SFMFeetToMeters(
            (i < argc) ? atof(argv[i]) : 80.0
        );

        /* Number of hazard lights? */
        i = 2;
        hazard_lights = (i < argc) ? atoi(argv[i]) : 0;
        flashing = 1;
        z = obj_premodel_ptr->height;
        for(i = 0; i < hazard_lights; i++)
        {
            light_ptr = SARObjCreateLight(
                scene,
                &obj_ptr->light, &obj_ptr->total_lights
            );
            if(light_ptr != NULL)
            {
                if(flashing)
                {
                    light_ptr->pos.x = 0.0;
                    light_ptr->pos.y = 0.0;
                    light_ptr->pos.z = z;

                    light_ptr->color.r = 1.0;
                    light_ptr->color.g = 0.0;
                    light_ptr->color.b = 0.0;
                    light_ptr->color.a = 1.0;

                    light_ptr->radius = 2;

                    light_ptr->flags |= SAR_LIGHT_FLAG_ON;
                    light_ptr->flags |= SAR_LIGHT_FLAG_STROBE;

                    light_ptr->int_on = 800;
                    light_ptr->int_off = 1700;
                }
                else
                {
                    light_ptr->pos.x = 0.0;
                    light_ptr->pos.y = 0.0;
                    light_ptr->pos.z = z;

                    light_ptr->color.r = 1.0;
                    light_ptr->color.g = 0.0;
                    light_ptr->color.b = 0.0;
                    light_ptr->color.a = 1.0;

                    light_ptr->radius = 2;

                    light_ptr->flags |= SAR_LIGHT_FLAG_ON;

                    light_ptr->int_on = 0;
                    light_ptr->int_off = 0;
                }
            }

            /* Toggle flashing, this is to create alternating steady
             * state and flashing hazard lights.
             */
            flashing = !flashing;

            /* Decrease height. */
            z -= (obj_premodel_ptr->height / hazard_lights);
        }

        /* Set cylendrical contact bounds, the radius of the contact
         * bounds is 0.3 times the height, since the height controls
         * a uniformed scaling the the unit width is 0.3.
         */
        SARObjAddContactBoundsCylendrical(
            obj_ptr,
            SAR_CRASH_FLAG_CRASH_CAUSE,
            SAR_CRASH_TYPE_OBSTRUCTION,
            5.0,                                /* Fixed radius. */
            0.0, obj_premodel_ptr->height       /* Height min and max. */
        );
}

/*
 *      Creates a tower.
 *
 *      Inputs assumed valid.
 */
void SARObjCreatePremodeledTower(
        sar_core_struct *core_ptr, sar_scene_struct *scene,
        int obj_num, sar_object_struct *obj_ptr,
        sar_object_premodeled_struct *obj_premodel_ptr,
        int argc, char **argv
)
{
        int flashing;
        double z;
        int i, hazard_lights;
        sar_light_struct *light_ptr;


        /* Format:
         *
         * <range> <height> <hazard_lights>
         */

        obj_premodel_ptr->type = SAR_OBJ_PREMODELED_TOWER;

        /* Range. */
        i = 0;
        obj_ptr->range = ((i < argc) ? atof(argv[i]) : 1000.0);

        /* Height (in feet). */
        i = 1;
        obj_premodel_ptr->height = SFMFeetToMeters(
            (i < argc) ? atof(argv[i]) : 80.0
        );

        /* Number of hazard lights? */
        i = 2;
        hazard_lights = (i < argc) ? atoi(argv[i]) : 0;
        flashing = 1;
        z = obj_premodel_ptr->height;
        for(i = 0; i < hazard_lights; i++)
        {
            light_ptr = SARObjCreateLight(
                scene,
                &obj_ptr->light, &obj_ptr->total_lights
            );
            if(light_ptr != NULL)
            {
                if(flashing)
                {
                    light_ptr->pos.x = 0.0;
                    light_ptr->pos.y = 0.0;
                    light_ptr->pos.z = z;

                    light_ptr->color.r = 1.0;
                    light_ptr->color.g = 0.0;
                    light_ptr->color.b = 0.0;
                    light_ptr->color.a = 1.0;

                    light_ptr->radius = 2;

                    light_ptr->flags |= SAR_LIGHT_FLAG_ON;
                    light_ptr->flags |= SAR_LIGHT_FLAG_STROBE;

                    light_ptr->int_on = 800;
                    light_ptr->int_off = 1700;
                }
                else
                {
                    light_ptr->pos.x = 0.0;
                    light_ptr->pos.y = 0.0;
                    light_ptr->pos.z = z;

                    light_ptr->color.r = 1.0;
                    light_ptr->color.g = 0.0;
                    light_ptr->color.b = 0.0;
                    light_ptr->color.a = 1.0;

                    light_ptr->radius = 2;

                    light_ptr->flags |= SAR_LIGHT_FLAG_ON;

                    light_ptr->int_on = 0;
                    light_ptr->int_off = 0;
                }
            }

            /* Toggle flashing, this is to create alternating steady
             * state and flashing hazard lights.
             */
            flashing = !flashing;

            /* Decrease height. */
            z -= (obj_premodel_ptr->height / hazard_lights);
        }

        /* Set cylendrical contact bounds, the radius of the contact
         * bounds is 0.3 times the height, since the height controls
         * a uniformed scaling the the unit width is 0.3.
         */
        SARObjAddContactBoundsCylendrical(
            obj_ptr,
            SAR_CRASH_FLAG_CRASH_CAUSE,
            SAR_CRASH_TYPE_OBSTRUCTION,
            5.0,				/* Fixed radius. */
            0.0, obj_premodel_ptr->height       /* Height min and max. */
        );
}

/*
 *	Creates a control tower.
 *
 *	Inputs assumed valid.
 */
void SARObjCreatePremodeledControlTower(
        sar_core_struct *core_ptr, sar_scene_struct *scene,
        int obj_num, sar_object_struct *obj_ptr,
        sar_object_premodeled_struct *obj_premodel_ptr,
        int argc, char **argv
)
{
        int i, n;

        /* Format:
         *
         * <range> <length> <width> <height> <walls_texture> <roof_texture>
         */

        obj_premodel_ptr->type = SAR_OBJ_PREMODELED_CONTROL_TOWER;

        /* Range. */
        i = 0;
        obj_ptr->range = ((i < argc) ? atof(argv[i]) : 1000.0);
            
        /* Length. */
        i = 1;
        obj_premodel_ptr->length = ((i < argc) ? atof(argv[i]) : 15.0);

        /* Width. */
        i = 2;
        obj_premodel_ptr->width = ((i < argc) ? atof(argv[i]) : 15.0);

        /* Height (in feet). */
        i = 3;
        obj_premodel_ptr->height = SFMFeetToMeters(
            (i < argc) ? atof(argv[i]) : 80.0
        );
  
        /* Walls texture. */
        i = 4;
        if(i < argc)
        {
            n = 0;
            if(n < SAR_OBJ_PREMODEL_MAX_TEXTURES)
                obj_premodel_ptr->tex_num[n] = SARGetTextureRefNumberByName(
                    scene, argv[i]
                );
        }

        /* Roof texture. */
        i = 5;
        if(i < argc)
        {
            n = 1;
            if(n < SAR_OBJ_PREMODEL_MAX_TEXTURES)
                obj_premodel_ptr->tex_num[n] = SARGetTextureRefNumberByName(
                    scene, argv[i]
                );
        }

        /* Set contact bounds. */
	SARObjAddContactBoundsRectangular(
	    obj_ptr,
	    SAR_CRASH_FLAG_CRASH_CAUSE | SAR_CRASH_FLAG_SUPPORT_SURFACE,
	    SAR_CRASH_TYPE_BUILDING,
	    -(obj_premodel_ptr->length / 2), (obj_premodel_ptr->length / 2),
	    -(obj_premodel_ptr->width / 2), (obj_premodel_ptr->width / 2),
	    0.0, obj_premodel_ptr->height
	);
}


/*
 *	Creates a building.
 *
 *	Inputs assumed valid.
 */
void SARObjCreatePremodeledBuilding(
	sar_core_struct *core_ptr, sar_scene_struct *scene,
	int obj_num, sar_object_struct *obj_ptr,
	sar_object_premodeled_struct *obj_premodel_ptr,
	int argc, char **argv
)
{
	int i, n;

	/* Format:
	 *
	 * <range> <length> <width> <height> <walls_texture>
	 * <walls_texture_night> <roof_texture>
	 */

	obj_premodel_ptr->type = SAR_OBJ_PREMODELED_BUILDING;

        /* Range. */
        i = 0;
	obj_ptr->range = ((i < argc) ? atof(argv[i]) : 1000.0);

	/* Length. */
	i = 1;
	obj_premodel_ptr->length = ((i < argc) ? atof(argv[i]) : 50.0);

	/* Width. */
	i = 2;
	obj_premodel_ptr->width = ((i < argc) ? atof(argv[i]) : 20.0);

	/* Height (in feet). */
	i = 3;
	obj_premodel_ptr->height = SFMFeetToMeters(
	    (i < argc) ? atof(argv[i]) : 50.0
	);

	/* Walls texture. */
	i = 4; n = 0;
	if(i < argc)
	{
	    if(n < SAR_OBJ_PREMODEL_MAX_TEXTURES)
	        obj_premodel_ptr->tex_num[n] = SARGetTextureRefNumberByName(
		    scene, argv[i]
	        );
	}

        /* Walls night. */
        i = 5; n = 1;
        if(i < argc)
        {
            if(n < SAR_OBJ_PREMODEL_MAX_TEXTURES)
                obj_premodel_ptr->tex_num[n] = SARGetTextureRefNumberByName(
                    scene, argv[i]
                );
        }

        /* Roof texture. */
        i = 6; n = 2;
        if(i < argc)
        {
            if(n < SAR_OBJ_PREMODEL_MAX_TEXTURES)
                obj_premodel_ptr->tex_num[n] = SARGetTextureRefNumberByName(
                    scene, argv[i]
                );
        }

        /* Set contact bounds. */
        SARObjAddContactBoundsRectangular(
            obj_ptr,
            SAR_CRASH_FLAG_CRASH_CAUSE | SAR_CRASH_FLAG_SUPPORT_SURFACE,
            SAR_CRASH_TYPE_BUILDING,
            -(obj_premodel_ptr->length / 2), (obj_premodel_ptr->length / 2),
            -(obj_premodel_ptr->width / 2), (obj_premodel_ptr->width / 2),
            0.0, obj_premodel_ptr->height
        );
}

/*
 *	Front end for creating a permodeled object.
 */
int SARObjCreatePremodeled(
	sar_core_struct *core_ptr, sar_scene_struct *scene,
	const char *type_name, int argc, char **argv
)
{
	int obj_num = -1;
	sar_object_struct *obj_ptr;
	sar_object_premodeled_struct *obj_premodel_ptr;


	if((core_ptr == NULL) || (scene == NULL) || (type_name == NULL))
	    return(obj_num);

	/* Create new premodeled object. */
	obj_num = SARObjCreate(
	    scene,
	    &core_ptr->object, &core_ptr->total_objects,
	    SAR_OBJ_TYPE_PREMODELED
        );
	if(obj_num < 0)
	    return(obj_num);
	else
 	    obj_ptr = core_ptr->object[obj_num];

	/* Get pointer to premodeled object data substructure. */
	obj_premodel_ptr = SAR_OBJ_GET_PREMODELED(obj_ptr);


	/* Handle by type name. */
	/* Standard building. */
	if(!strcasecmp(type_name, "building"))
	{
	    SARObjCreatePremodeledBuilding(
		core_ptr, scene,
		obj_num, obj_ptr, obj_premodel_ptr,
		argc, argv
	    );
	}
	/* Control tower. */
	else if(!strcasecmp(type_name, "control_tower"))
        {
	    SARObjCreatePremodeledControlTower(
                core_ptr, scene,
                obj_num, obj_ptr, obj_premodel_ptr,
                argc, argv
            );
	}
        /* Tower. */
        else if(!strcasecmp(type_name, "tower"))
        {
            SARObjCreatePremodeledTower(
                core_ptr, scene,
                obj_num, obj_ptr, obj_premodel_ptr,
                argc, argv
            );
        }
        /* Radio tower. */
        else if(!strcasecmp(type_name, "radio_tower"))
        {
            SARObjCreatePremodeledRadioTower(
                core_ptr, scene,
                obj_num, obj_ptr, obj_premodel_ptr,
                argc, argv
            );
        }
        /* Power transmission tower. */
        else if(!strcasecmp(type_name, "power_transmission_tower"))
        {
            SARObjCreatePremodeledPowerTransmissionTower(
                core_ptr, scene,
                obj_num, obj_ptr, obj_premodel_ptr,
                argc, argv
            );
        }

	else
	{
	    fprintf(
		stderr,
		"Unsupported permodeled type name `%s'.\n",
		type_name
	    );
	}

	return(obj_num);
}
