#include <stdio.h>
#include <sys/types.h>
#include <math.h>

#include "../include/os.h"

#ifdef __MSW__
# include <windows.h>
#endif

#include <GL/gl.h>
#include <GL/glu.h>

#include "gw.h"

#include "sarreality.h"
#include "obj.h"
#include "sfm.h"
#include "sar.h"
#include "sardrawpm.h"


void SARDrawPremodeledPowerTransmissionTower(
        gw_display_struct *display,
        sar_scene_struct *scene,
        sar_object_struct *obj_ptr,
        sar_object_premodeled_struct *obj_premodeled_ptr,
        double camera_dist,
        Boolean draw_for_gcc
);


#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)     (MIN(MAX((a),(l)),(h)))


/*
 *      Draws a premodeled power transmission tower.
 */
void SARDrawPremodeledPowerTransmissionTower(
        gw_display_struct *display,
        sar_scene_struct *scene,
        sar_object_struct *obj_ptr,
        sar_object_premodeled_struct *obj_premodeled_ptr,
        double camera_dist,
        Boolean draw_for_gcc
)
{
        int i;
        GLenum shade_model_mode;
        double sx, sy, sz;              /* Scale. */


        /* All hard coded coordinates will produce an object of unit
         * size (height = 1.0). So the sx, sy, and sz scale coefficients
         * will simply multiply each point.
         */

        /* Calculate scales, all based on height. */
        sx = sy = sz = obj_premodeled_ptr->height;

        /* Set flat shade model state. */
        shade_model_mode = display->state_gl.shade_model_mode;
        StateGLShadeModel(&display->state_gl, GL_FLAT);

        glColor4d(0.45, 0.45, 0.45, 1.0);

        /* Draw close range model? */
        if(camera_dist < 2000.0)
        {
          /* Draw two passes (one for the north side, the other for
           * the south side.
           */
          for(i = 0; i < 2; i++)
          {
            glBegin(GL_LINE_STRIP);
            {
                /* North base. */
/*              glNormal3d(0, 0.2, -(0.98 * ((sy < 0.0) ? -1.0 : 1.0))); */
                glNormal3d(0.0, 1.0, 0.0);

                glVertex3d(-0.12 * sx, 0.0 * sz, -(0.12 * sy));
                glVertex3d(-0.08 * sx, 0.2 * sz, -(0.08 * sy));
                glVertex3d(0.08 * sx, 0.2 * sz, -(0.08 * sy));
                glVertex3d(0.12 * sx, 0.0 * sz, -(0.12 * sy));
                glVertex3d(-0.08 * sx, 0.2 * sz, -(0.08 * sy));
            }
            glEnd();
            glBegin(GL_LINE_STRIP);
            {
                glVertex3d(-0.12 * sx, 0.0 * sz, -(0.12 * sy));
                glVertex3d(0.08 * sx, 0.2 * sz, -(0.08 * sy));
                glVertex3d(-0.05 * sx, 0.4 * sz, -(0.05 * sy));
                glVertex3d(0.05 * sx, 0.4 * sz, -(0.05 * sy));
                glVertex3d(-0.08 * sx, 0.2 * sz, -(0.08 * sy));
            }
            glEnd();
            glBegin(GL_LINE_STRIP);
            {
                glVertex3d(-0.08 * sx, 0.2 * sz, -(0.08 * sy));
                glVertex3d(-0.05 * sx, 0.4 * sz, -(0.05 * sy));
                glVertex3d(0.05 * sx, 0.65 * sz, -(0.05 * sy));
                glVertex3d(0.05 * sx, 0.4 * sz, -(0.05 * sy));
            }
            glEnd();
            glBegin(GL_LINE_STRIP);
            {
                glVertex3d(-0.05 * sx, 0.4 * sz, -(0.05 * sy));
                glVertex3d(-0.05 * sx, 0.65 * sz, -(0.05 * sy));
                glVertex3d(0.05 * sx, 0.4 * sz, -(0.05 * sy));
                glVertex3d(0.08 * sx, 0.2 * sz, -(0.02 * sy));
            }
            glEnd();
            glBegin(GL_LINE_STRIP);
            {
                glVertex3d(0.05 * sx, 0.65 * sz, -(0.05 * sy));
                glVertex3d(-0.05 * sx, 0.65 * sz, -(0.05 * sy));
                glVertex3d(-0.05 * sx, 0.9 * sz, -(0.05 * sy));
                glVertex3d(0.05 * sx, 0.65 * sz, -(0.05 * sy));
                glVertex3d(0.05 * sx, 0.9 * sz, -(0.05 * sy));
            }
            glEnd();
            glBegin(GL_LINE_STRIP);
            {
                glVertex3d(-0.05 * sx, 0.65 * sz, -(0.05 * sy));
                glVertex3d(0.05 * sx, 0.9 * sz, -(0.05 * sy));
                glVertex3d(0.0 * sx, 1.0 * sz, -(0.0 * sy));
                glVertex3d(-0.05 * sx, 0.9 * sz, -(0.05 * sy));
                glVertex3d(0.05 * sx, 0.9 * sz, -(0.05 * sy));
            }
            glEnd();
            glBegin(GL_LINE_STRIP);
            {
                glVertex3d(-0.05 * sx, 0.65 * sz, -(0.05 * sy));
                glVertex3d(-0.25 * sx, 0.65 * sz, -(0.05 * sy));
                glVertex3d(-0.27 * sx, 0.56 * sz, -(0.0 * sy));
                glVertex3d(-0.22 * sx, 0.62 * sz, -(0.05 * sy));
                glVertex3d(-0.05 * sx, 0.59 * sz, -(0.05 * sy));
                glVertex3d(-0.25 * sx, 0.65 * sz, -(0.05 * sy));
                glVertex3d(-0.22 * sx, 0.62 * sz, -(0.05 * sy));
                glVertex3d(-0.05 * sx, 0.65 * sz, -(0.05 * sy));
            }
            glEnd();
            glBegin(GL_LINE_STRIP);
            {
                glVertex3d(0.05 * sx, 0.65 * sz, -(0.05 * sy));
                glVertex3d(0.25 * sx, 0.65 * sz, -(0.05 * sy));
                glVertex3d(0.27 * sx, 0.56 * sz, -(0.0 * sy));
                glVertex3d(0.22 * sx, 0.62 * sz, -(0.05 * sy));
                glVertex3d(0.05 * sx, 0.59 * sz, -(0.05 * sy));
                glVertex3d(0.25 * sx, 0.65 * sz, -(0.05 * sy));
                glVertex3d(0.22 * sx, 0.62 * sz, -(0.05 * sy));
                glVertex3d(0.05 * sx, 0.65 * sz, -(0.05 * sy));
            }
            glEnd();
            glBegin(GL_LINE_STRIP);
            {
                glVertex3d(-0.05 * sx, 0.9 * sz, -(0.05 * sy));
                glVertex3d(-0.23 * sx, 0.9 * sz, -(0.05 * sy));
                glVertex3d(-0.25 * sx, 0.81 * sz, -(0.0 * sy));
                glVertex3d(-0.2 * sx, 0.87 * sz, -(0.05 * sy));
                glVertex3d(-0.05 * sx, 0.84 * sz, -(0.05 * sy));
                glVertex3d(-0.23 * sx, 0.9 * sz, -(0.05 * sy));
                glVertex3d(-0.2 * sx, 0.87 * sz, -(0.05 * sy));
                glVertex3d(-0.05 * sx, 0.9 * sz, -(0.05 * sy));
            }
            glEnd();
            glBegin(GL_LINE_STRIP);
            {
                glVertex3d(0.05 * sx, 0.9 * sz, -(0.05 * sy));
                glVertex3d(0.23 * sx, 0.9 * sz, -(0.05 * sy));
                glVertex3d(0.25 * sx, 0.81 * sz, -(0.0 * sy));
                glVertex3d(0.2 * sx, 0.87 * sz, -(0.05 * sy));
                glVertex3d(0.05 * sx, 0.84 * sz, -(0.05 * sy));
                glVertex3d(0.23 * sx, 0.9 * sz, -(0.05 * sy));
                glVertex3d(0.2 * sx, 0.87 * sz, -(0.05 * sy));
                glVertex3d(0.05 * sx, 0.9 * sz, -(0.05 * sy));
            }
            glEnd();

            /* Flip the y scale value to draw the south side. */
            sy *= -1.0;
          }
        }
        else
        {
            /* Draw far model. */
            glBegin(GL_LINES);
            {
                glNormal3d(0.0, 1.0, 0.0);

                glVertex3d(0.0 * sx, 0.0 * sz, -(0.0 * sy));
                glVertex3d(0.0 * sx, 1.0 * sz, -(0.0 * sy));

                glVertex3d( 0.27 * sx, 0.62 * sz, -(0.0 * sy));
                glVertex3d(-0.27 * sx, 0.62 * sz, -(0.0 * sy));

                glVertex3d( 0.26 * sx, 0.9 * sz, -(0.0 * sy));
                glVertex3d(-0.26 * sx, 0.9 * sz, -(0.0 * sy));
            }
            glEnd();
        }

        /* Reget y scale, since it has been modified in the above loop. */
        sy = obj_premodeled_ptr->height;


        /* Restore shade model. */
        StateGLShadeModel(&display->state_gl, shade_model_mode);
}
