#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <sys/types.h>

#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>

#define pixel_t	unsigned long


void DoWriteFile(
	char *path,
	u_int8_t *bitmap,
	int width, int height,
	char c
)
{
	int i, y, n;
	int bytes_per_line;
	FILE *fp;


	if(bitmap == NULL)
	    return;

	fp = fopen(path, "a");
	if(fp == NULL)
	    return;

	if((width % 8) == 0)
	    bytes_per_line = width / 8;
	else
	    bytes_per_line = (width / 8) + 1;

	fprintf(fp, "/* Character %i */\n", (unsigned char)c);

	/* Save backwards. */
	for(y = height - 1, n = 0; y >= 0; y--)
	{
	    for(i = 0; i < bytes_per_line; i++)
	    {
	        if(n >= 8)
	        {
		    fputc('\n', fp);
		    n = 0;
	        }
	        fprintf(fp, "0x%.2x, ",
		    bitmap[
			(bytes_per_line * y) +
			i
		    ]
		);
		n++;
	    }
	}

	/* Seek back two characters. */
	fseek(fp, -2, SEEK_CUR);

        fprintf(fp, "\n");

	fclose(fp);


	return;
}


u_int8_t *DoCreateBitmap(
	u_int8_t *data,
	int bytes_per_line,
	int width, int height,
	int depth
)
{
	u_int8_t *ptr8, *bm_ptr8;
	int n;
        int bm_bytes_per_line;
/*	int bm_byte_pos; */
	int bm_bytes_len;

	u_int8_t *img_ptr8;
	int x, y, bytes_per_pixel;


	if(data == NULL)
	    return(NULL);

	/* Allocate bitmap data. */
	if((width % 8) == 0)
	    bm_bytes_per_line = width / 8;
	else
	    bm_bytes_per_line = (width / 8) + 1;
	if(bm_bytes_per_line < 1)
	    bm_bytes_per_line = 1;

	bm_bytes_len = bm_bytes_per_line * height;
	ptr8 = (u_int8_t *)calloc(bm_bytes_len, sizeof(u_int8_t));
	if(ptr8 == NULL)
	    return(NULL);

	switch(depth)
	{
	  case 8: bytes_per_pixel = 1; break;
	  case 15: case 16: bytes_per_pixel = 2; break;
	  case 24: case 32: bytes_per_pixel = 4; break;
	  default: bytes_per_pixel = 1; break;
	}

	/* Go through each pixel on each line. */
	for(y = 0; y < height; y++)
	{
	    for(x = 0; x < width; x++)
	    {
		/* Get pointer to image pixel. */
		img_ptr8 = (u_int8_t *)(&data[
		    (bytes_per_line * y) +
                    (bytes_per_pixel * x)
		]);
		if(*img_ptr8 == 0x00)
		    continue;

		/* Get pointer to lowest byte in bitmap. */
		bm_ptr8 = (u_int8_t *)(&ptr8[
		    (bm_bytes_per_line * y) +
		    (x / 8)
		]);

		n = 7 - (x % 8);

		*bm_ptr8 += (1 << n);
	    }
	}

	return(ptr8);
}

void GenerateFont(
	Display *dpy, GC gc, Window root_win, int depth,
	int offset_x, int offset_y, int font_width, int font_height,
	XFontStruct *font, char c,
	pixel_t white_pix, pixel_t black_pix,
	char *output_file
)
{
	XImage *image;
	Pixmap pixmap;
        u_int8_t *ptr8;
	char string[2];


	/* Create a Pixmap to draw to. */
	pixmap = XCreatePixmap(
	    dpy, root_win,
	    font_width, font_height, depth
	);
	if(pixmap == None)
	    return;

	/* Clear pixmap. */
	XSetForeground(dpy, gc, black_pix);
	XFillRectangle(
	    dpy, pixmap, gc,
	    0, 0, font_width, font_height
	);


        XSetForeground(dpy, gc, white_pix);
        XSetFont(dpy, gc, font->fid);

	string[0] = c;
	string[1] = '\0';
	/* Is there per character data? This implies font is variable
	 * width and we need to center each character.
	 */
	if(font->per_char != NULL)
        {
            XCharStruct *csp;

            /* 8 bits per character type? */
            if(!font->min_byte1 && !font->max_byte1)
            {
		int ci = c - (int)font->min_char_or_byte2;	/* Character value index. */
                int i = 0;
                int m = (int)font->max_char_or_byte2 -
                        (int)font->min_char_or_byte2;

		/* Clip, check if string char in bounds? */
		if((ci >= i) && (ci <= m))
		{
		    int this_char_width;

		    csp = &font->per_char[ci];       /* Ptr to geometry. */

		    /* Recalculate offset_x. */
		    this_char_width = csp->rbearing - csp->lbearing;
		    offset_x = (font_width / 2) - (this_char_width / 2)
			- csp->lbearing;
		}
	    }
	}
	XDrawString(dpy, pixmap, gc, offset_x, offset_y, string, 1);

	XSync(dpy, False);

	image = XGetImage(dpy, pixmap, 0, 0, font_width, font_height,
	    AllPlanes, ZPixmap
	);
	if(image == NULL)
	{
	    XFreePixmap(dpy, pixmap);
	    return;
	}

	ptr8 = DoCreateBitmap(
	    (u_int8_t *)image->data,
	    image->bytes_per_line,
	    image->width,
	    image->height,
	    depth
	);
	if(ptr8 != NULL)
	{
	    DoWriteFile(output_file, ptr8, font_width, font_height, c);
	    free(ptr8);
	}

	XDestroyImage(image);

/*
	w = XCreateSimpleWindow(
	    dpy, root_win, 0, 0,
	    font_width, font_height,
	    0, 0, 0
	);
	if(w != None)
	{
	    XMapRaised(dpy, w);
	    XSync(dpy, False);

	    XCopyArea(
		dpy, pixmap, w, gc,
		0, 0,
		font_width, font_height,
		0, 0
	    );
	    XSync(dpy, False);

	    sleep(5);
	}
 */

	XFreePixmap(dpy, pixmap);

	return;
}

int main(int argc, char **argv)
{
	int i;
	FILE *fp;
	Display *dpy;
	int scr_num;
	pixel_t white_pix, black_pix;
	GC gc;
	int depth;
	Window root_win;
	XFontStruct *font;
	int font_width, font_height, offset_x, offset_y;
	char *fontname = "8x16";
	char *output_file = "output.fnt";


	/* Parse arguments. */
	if(argc < 5)
	{
	    printf("Usage: %s width height fn output\n", argv[0]);
	    return(0);
	}

        i = 1;
        font_width = atoi((argc > i) ? argv[i] : "0");
        i = 2;
        font_height = atoi((argc > i) ? argv[i] : "0");
	i = 3;
	fontname = ((argc > i) ? argv[i] : "8x16");
        i = 4;
        output_file = ((argc > i) ? argv[i] : "output.fnt");


	dpy = XOpenDisplay(NULL);
	if(dpy == NULL)
	{
	    fprintf(stderr,
 "Cannot connect to X server.\n"
	    );
	    return(1);
	}

	scr_num = DefaultScreen(dpy);
        white_pix = WhitePixel(dpy, scr_num);
        black_pix = BlackPixel(dpy, scr_num);
        depth = DefaultDepth(dpy, scr_num);
	root_win = DefaultRootWindow(dpy);
	gc = DefaultGC(dpy, scr_num);  


	/* Load font. */
	font = XLoadQueryFont(dpy, fontname);
	if(font == NULL)
	{
	    fprintf(stderr,
		"%s: No such font.\n",
		fontname
	    );

	    XCloseDisplay(dpy);
	    return(0);
	}

	/* Set offsets (upper left corner oriented). */
	offset_x = font->min_bounds.lbearing;
	offset_y = font->max_bounds.ascent + (font->max_bounds.descent / -2);

	/* Get font width and heights as needed. */
	/* The font width is just the max bounds. */
	if(font_width <= 0)
	    font_width = font->max_bounds.width;
	/* Font height needs to be increased by half of the descent for
	 * proper centering on result.
	 */
	if(font_height <= 0)
	    font_height = font->max_bounds.ascent +
		font->max_bounds.descent + (font->max_bounds.descent / -2);

printf("Generating bitmaps for font `%s'\n", fontname);
printf("Font size bounds: %ix%i\n", font_width, font_height);


	/* Write header. */
	fp = fopen(output_file, "w");
	if(fp != NULL)
	{
	    int m = 0;

	    if((font_width % 8) == 0)
		m = font_width / 8;
	    else
		m = (font_width / 8) + 1;

	    fprintf(fp, "static unsigned char test_font[] = {\n");
            fprintf(fp, "/* First 32 bytes are header. */\n");

            fprintf(fp,
"0x%.2x, 0x%.2x, 0x%.2x, 0x%.2x, 0x%.2x, 0x00, 0x00, 0x00,\n",
		(unsigned char)font_width,
		(unsigned char)font_height,
		(unsigned char)font_width + 2,
		(unsigned char)font_height + 2,
		(unsigned char)m
	    );
            fprintf(fp,
"0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,\n"
	    );
            fprintf(fp,
"0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,\n"
            );
            fprintf(fp,
"0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,\n"
            );

	    fclose(fp);
	}

	/* Generate fonts. */
	for(i = 0; i < 256; i++)
	{
printf("\r Rendering character %i of %i... ", i + 1, 256);
fflush(stdout);
	    GenerateFont(
	        dpy, gc, root_win, depth,
	        offset_x, offset_y, font_width, font_height,
	        font, (char)i, white_pix, black_pix, output_file
	    );
	}
printf("\nDone.\n");

        /* Write footer. */
        fp = fopen(output_file, "a+");
        if(fp != NULL)
        {
	    fseek(fp, -2, SEEK_CUR);

	    fprintf(fp, "\n};\n");

	    fclose(fp);
	}


	XFreeFont(dpy, font);
        XCloseDisplay(dpy);

	return(0);
}
